from collections import namedtuple
import torch.nn as nn
from .utils import _isinstance

Config = namedtuple('config', [
    'conv_and_linear_weight', 'conv_and_linear_bias', 'bn_weight', 'bn_bias',
    'acti'
],
                    defaults=[8, 8, 8, 8, 8])


def config_bn(proc, config, bn_list):
    if _isinstance(proc, bn_list):
        proc.weight_bit_width = config.bn_weight
        proc.bias_bit_width = config.bn_bias


def config_acti(proc, config, acti_list):
    if _isinstance(proc, acti_list):
        proc.acti_bit_width = config.acti


def config_conv_linear(proc, config, conv_linear_list):
    if _isinstance(proc, conv_linear_list):
        proc.weight_bit_width = config.conv_and_linear_weight
        proc.bias_bit_width = config.conv_and_linear_bias


def config_network(
        net,
        name,
        config,
        bn_list=[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d],
        acti_list=[nn.ReLU, nn.ReLU6],
        conv_linear_list=[nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.Linear],
        show=False):
    proc_list = list(net._modules.keys())
    if proc_list == []:
        if show:
            print(name, 'is configured')
        config_acti(net, config, acti_list=acti_list)
        config_conv_linear(net, config, conv_linear_list=conv_linear_list)
        config_bn(net, config, bn_list=bn_list)
    else:
        for n in proc_list:
            config_network(net._modules[n],
                           name + '.' + n,
                           config,
                           bn_list=bn_list,
                           acti_list=acti_list,
                           conv_linear_list=conv_linear_list,
                           show=show)



import torch
import torch.nn as nn
from collections import OrderedDict
from .convbnact import Conv2dBNReLU


def fold_op(conv, bn, relu):
    folded = Conv2dBNReLU(conv, bn, relu)
    conv = folded
    bn = nn.Identity()
    relu = nn.Identity()
    return conv, bn, relu


def fold_the_network(net):
    key = list(net._modules.keys())
    keylen = len(key)
    flag = 0
    while flag < keylen:
        if list(net._modules[key[flag]]._modules.keys()) != []:
            fold_the_network(net._modules[key[flag]])
        else:
            if isinstance(net._modules[key[flag]],
                          nn.Conv2d) and flag + 1 < keylen:
                if isinstance(net._modules[key[flag + 1]],
                              nn.BatchNorm2d) and flag + 2 < keylen:
                    if isinstance(net._modules[key[flag + 2]],
                                  nn.ReLU) or isinstance(
                                      net._modules[key + 2], nn.ReLU6):
                        net._modules[key[flag]], net._modules[key[
                            flag + 1]], net._modules[key[flag + 2]] = fold_op(
                                net._modules[key[flag]],
                                net._modules[key[flag + 1]],
                                net._modules[key[flag + 2]])
                        print(flag)
                        flag += 2
        flag += 1



"""
Provide quantilized form of torch.nn.modules.activation
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import qsigned, qunsigned


class ReLU(nn.ReLU):
    def __init__(self,
                 inplace=False,
                 acti_bit_width=8,
                 retrain=True,
                 quant=False):
        super().__init__(inplace)
        self.acti_bit_width = acti_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.acti_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.acti_log2_t = torch.Tensor(1)

    def relu_forward(self, input):
        return qunsigned(F.relu(input), self.acti_log2_t, self.acti_bit_width)

    def relu_forward_unquant(self, input):
        return F.relu(input)

    def quantilize(self):
        self.quant = True
        self.acti_log2_t.requires_grad = False

    def floatilize(self):
        self.quant = False
        self.acti_log2_t.requires_grad = True

    def forward(self, input):
        return self.relu_forward(
            input) if self.quant else self.relu_forward_unquant(input)


class ReLU6(nn.ReLU6):
    def __init__(self,
                 inplace=False,
                 acti_bit_width=8,
                 retrain=True,
                 quant=False):
        super().__init__(inplace)
        self.acti_bit_width = acti_bit_width
        self.retrain = retrain
        self.quant = quant
        if retrain is True:
            self.acti_log2_t = nn.Parameter(torch.Tensor(1))
        else:
            self.acti_log2_t = torch.Tensor(1)

    def relu6_forward(self, input):
        return qunsigned(F.relu6(input), self.acti_log2_t, self.acti_bit_width)

    def relu6_forward_unquant(self, input):
        return F.relu6(input)

    def quantilize(self):
        self.quant = True

    def floatilize(self):
        self.quant = False

    def forward(self, input):
        return self.relu6_forward(
            input) if self.quant else self.relu6_forward_unquant(input)



import torch
import torch.nn as nn
from ..function import qsigned, qunsigned
from ..threshold import add_hook, hook_handler, get_hook


def kl_divergence(ha, hb):
    r'''
        ans = J_{kl}(a,b)
    '''
    return (ha * (ha / hb).log()).sum()


def kl_similarity(
        net,
        qnet,
        cali_batch,
        end_list=[nn.Conv2d, nn.Linear, nn.ReLU, nn.BatchNorm2d, nn.ReLU6],
        show=True):
    ret = []
    add_hook(net, 'net', hook_handler, end_list=end_list, show=True)
    add_hook(qnet, 'net', hook_handler, end_list=end_list, show=True)
    net(cali_batch)
    qnet(cali_batch)
    n_hook = get_hook(net, '', show=True)
    q_hook = get_hook(qnet, '', show=True)
    print(n_hook)
    for n, q in zip(n_hook, q_hook):
        kl = kl_divergence(n[1].flatten(), q[1].flatten())
        ret.append((n[0], kl))
        if show:
            print(f'kl at {n[0]} is {kl}, average is {kl/n[1].numel()}')
    return ret


import torch.nn as nn
from ..utils import _isinstance


def hook_handler(model, input, output):
    model.hook_out = output


def remove_hook(handles):
    for handle in handles:
        handle.remove()


def remove_hookout(layer, name, show=False):
    keys = list(layer._modules.keys())
    if keys == []:
        if show:
            print(name, 'has removed hookout')
        delattr(layer, 'hook_out')
    else:
        for key in keys:
            remove_hookout(layer._modules[key], name + '.' + key, show=show)


def add_hook(
        layer,
        name,
        hook_handler,
        end_list=[nn.Conv2d, nn.Linear, nn.ReLU, nn.BatchNorm2d, nn.ReLU6],
        show=False):
    handles = []
    keys = list(layer._modules.keys())
    if _isinstance(layer, end_list):
        if show:
            print(name, 'has added a hook')
        handles.append(layer.register_forward_hook(hook_handler))
    for key in keys:
        handles.extend(
            add_hook(layer._modules[key],
                     name + '.' + key,
                     hook_handler,
                     end_list=end_list,
                     show=show))
    return handles


def get_hook(layer, name, show=False):
    hooks_got = []
    keys = list(layer._modules.keys())
    if hasattr(layer, 'hook_out'):
        hooks_got.append((name, layer.hook_out))
        if show:
            print(name, 'hook has been fetched')
    for key in keys:
        hooks_got.extend(
            get_hook(layer._modules[key], name + '.' + key, show=show))
    return hooks_got



import torch
import torch.nn as nn


def threshold_weight_ig(module, qmodule, eps=1e-8):
    ig_value = torch.tensor([0.])
    qmodule.weight_log2_t = torch.nn.Parameter(
        ig_value) if qmodule.retrain else ig_value


def threshold_bias_ig(module, qmodule, eps=1e-8):
    ig_value = torch.tensor([qmodule.bias_bit_width / 2.]) + eps
    qmodule.bias_log2_t = torch.nn.Parameter(
        ig_value) if qmodule.retrain else ig_value


def threshold_activation_ig(module, qmodule, eps=1e-8):
    ig_value = torch.tensor([qmodule.acti_bit_width - 1.]) + eps
    qmodule.acti_log2_t = torch.nn.Parameter(
        ig_value) if qmodule.retrain else ig_value



import torch

num_list = [str(i) for i in range(0, 10)]


def _getattr(obj, key):
    stage = key.rsplit('.')
    for i in stage:
        if i[0] in num_list:
            obj = obj[int(i)]
        else:
            obj = getattr(obj, i)
    return obj


def _isinstance(obj, cls_list):
    flag_list = [isinstance(obj, cls) for cls in cls_list]
    flag = False
    for x in flag_list:
        flag = x or flag
    return flag


def make_net_quant_or_not(net_proc, name, quant=True, exclude=[], show=False):
    keys = list(net_proc._modules.keys())
    if hasattr(net_proc, 'quant'):
        if not _isinstance(net_proc, exclude):
            getattr(net_proc, 'quantilize')() if quant else getattr(
                net_proc, 'floatilize')()
            if show:
                print(name, ' is quanted')
        else:
            getattr(net_proc, 'floatilize')()
            if show:
                print(name, 'is excluded')
    for key in keys:
        make_net_quant_or_not(net_proc._modules[key],
                              name + f'.{key}',
                              quant=quant,
                              exclude=exclude,
                              show=show)


def make_bn_fold_with_previous(net_proc, name, pre_proc, pre_name, show=False):
    keys = list(net_proc._modules.keys())
    _pre_proc = None
    _pre_name = ''
    if isinstance(net_proc, torch.nn.BatchNorm2d):
        if pre_proc is not None:
            net_proc.fold(pre_proc)
            if show:
                print(f'{name} is folded with {pre_name}')
    return net_proc, name
    for key in keys:
        _pre_proc, _pre_name = make_bn_fold_with_previous(
            net_proc._modules[key],
            name + f'.{key}',
            _pre_proc,
            _pre_name,
            show=show)
    return _pre_proc, _pre_name



import torch
from ..function import qsigned, qunsigned
from collections import OrderedDict


def cosine_similarity(net, qnet, wba=[8, 16, 8], show=False):
    ret = []
    n_param, q_param = OrderedDict(net.named_parameters()), OrderedDict(
        qnet.named_parameters())
    for name in n_param.keys():
        if name.find('log') == -1:
            qparam = q_param[name]
            if name.find('weight') != -1:
                width = wba[0]
            elif name.find('bias'):
                width = wba[1]
            else:
                width = wba[2]
            param = n_param[name]
            qparam = qsigned(qparam, q_param[name + '_log2_t'], width)
            cos = param.flatten().dot(
                qparam.flatten()) / param.norm() / qparam.norm()
            ret.append((name, cos))
            if show:
                print(f'layer {name} get cosine of {cos}')
